<?php
/**
 * A general file class
 *
 * @author		Roland Haeder <webmaster@ship-simu.org>
 * @version		0.0.0
 * @copyright	Copyright (c) 2007, 2008 Roland Haeder, 2009 - 2012 Core Developer Team
 * @license		GNU GPL 3.0 or any newer version
 * @link		http://www.ship-simu.org
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
class BaseFile extends BaseFrameworkSystem {
	/**
	 * The current file we are working in
	 */
	private $fileName = '';

	/**
	 * Back-buffer
	 */
	private $backBuffer = '';

	/**
	 * Currently loaded block (will be returned by current())
	 */
	private $currentBlock = '';

	/**
	 * Protected constructor
	 *
	 * @param	$className	Name of the class
y	 * @return	void
	 */
	protected function __construct ($className) {
		// Call parent constructor
		parent::__construct($className);

		// Init counters and gaps array
		$this->initCountersGapsArray();
	}

	/**
	 * Destructor for cleaning purposes, etc
	 *
	 * @return	void
	 */
	public final function __destruct() {
		// Try to close a file
		$this->closeFile();

		// Call the parent destructor
		parent::__destruct();
	}

	/**
	 * Getter for the file pointer
	 *
	 * @return	$filePointer	The file pointer which shall be a valid file resource
	 * @throws	UnsupportedOperationException	If this method is called
	 */
	public final function getPointer () {
		throw new UnsupportedOperationException(array($this, __FUNCTION__), self::EXCEPTION_UNSPPORTED_OPERATION);
	}

	/**
	 * Setter for file name
	 *
	 * @param	$fileName	The new file name
	 * @return	void
	 */
	protected final function setFileName ($fileName) {
		$fileName = (string) $fileName;
		$this->fileName = $fileName;
	}

	/**
	 * Getter for file name
	 *
	 * @return	$fileName	The current file name
	 */
	public final function getFileName () {
		return $this->fileName;
	}

	/**
	 * Initializes the back-buffer by setting it to an empty string.
	 *
	 * @return	void
	 */
	private function initBackBuffer () {
		// Simply call the setter
		$this->setBackBuffer('');
	}

	/**
	 * Setter for backBuffer field
	 *
	 * @param	$backBuffer		Characters to "store" in back-buffer
	 * @return	void
	 */
	private function setBackBuffer ($backBuffer) {
		// Cast to string (so no arrays or objects)
		$backBuffer = (string) $backBuffer;

		// ... and set it
		$this->backBuffer = $backBuffer;
	}

	/**
	 * Getter for backBuffer field
	 *
	 * @return	$backBuffer		Characters "stored" in back-buffer
	 */
	private function getBackBuffer () {
		return $this->backBuffer;
	}

	/**
	 * Setter for currentBlock field
	 *
	 * @param	$currentBlock		Characters to set a currently loaded block
	 * @return	void
	 */
	private function setCurrentBlock ($currentBlock) {
		// Cast to string (so no arrays or objects)
		$currentBlock = (string) $currentBlock;

		// ... and set it
		$this->currentBlock = $currentBlock;
	}

	/**
	 * Gets currently read data
	 *
	 * @return	$current	Currently read data
	 */
	public function getCurrentBlock () {
		// Return it
		return $this->currentBlock;
	}

	/**
	 * Initializes this file class
	 *
	 * @param	$fileName	Name of this abstract file
	 * @return	void
	 */
	protected function initFile ($fileName) {
		// Get a file i/o pointer instance
		$pointerInstance = ObjectFactory::createObjectByConfiguredName('file_raw_input_output_class', array($fileName));

		// ... and set it here
		$this->setPointerInstance($pointerInstance);
	}

	/**
	 * Close a file source and set it's instance to null and the file name
	 * to empty
	 *
	 * @return	void
	 * @todo	~10% done?
	 */
	public function closeFile () {
		$this->partialStub('Unfinished method.');

		// Remove file name
		$this->setFileName('');
	}

	/**
	 * Determines seek position
	 *
	 * @return	$seekPosition	Current seek position
	 */
	public function determineSeekPosition () {
		// Call pointer instance
		return $this->getPointerInstance()->determineSeekPosition();
	}

	/**
	 * Seek to given offset (default) or other possibilities as fseek() gives.
	 *
	 * @param	$offset		Offset to seek to (or used as "base" for other seeks)
	 * @param	$whence		Added to offset (default: only use offset to seek to)
	 * @return	$status		Status of file seek: 0 = success, -1 = failed
	 */
	public function seek ($offset, $whence = SEEK_SET) {
		// Call pointer instance
		return $this->getPointerInstance()->seek($offset, $whence);
	}

	/**
	 * Size of this file
	 *
	 * @return	$size	Size (in bytes) of file
	 * @todo	Handle seekStatus
	 */
	public function size () {
		// Call pointer instance
		return $this->getPointerInstance()->size();
	}

	/**
	 * Read data a file pointer
	 *
	 * @return	mixed	The result of fread()
	 * @throws	NullPointerException	If the file pointer instance
	 *									is not set by setPointer()
	 * @throws	InvalidResourceException	If there is being set
	 */
	public function readFromFile () {
		// Call pointer instance
		return $this->getPointerInstance()->readFromFile();
	}

	/**
	 * Reads given amount of bytes from file.
	 *
	 * @param	$bytes	Amount of bytes to read
	 * @return	$data	Data read from file
	 */
	public function read ($bytes) {
		// Call pointer instance
		return $this->getPointerInstance()->read($bytes);
	}

	/**
	 * Write data to a file pointer
	 *
	 * @param	$dataStream		The data stream we shall write to the file
	 * @return	mixed			Number of writes bytes or FALSE on error
	 * @throws	NullPointerException	If the file pointer instance
	 *									is not set by setPointer()
	 * @throws	InvalidResourceException	If there is being set
	 *											an invalid file resource
	 */
	public function writeToFile ($dataStream) {
		// Call pointer instance
		return $this->getPointerInstance()->writeToFile($dataStream);
	}

	/**
	 * Rewinds to the beginning of the file
	 *
	 * @return	$status 	Status of this operation
	 */
	public function rewind () {
		// Call pointer instance
		return $this->getPointerInstance()->rewind();
	}

	/**
	 * Determines whether the EOF has been reached
	 *
	 * @return	$isEndOfFileReached		Whether the EOF has been reached
	 */
	public final function isEndOfFileReached () {
		// Call pointer instance
		return $this->getPointerInstance()->isEndOfFileReached();
	}

	/**
	 * Analyzes entries in index file. This will count all found (and valid)
	 * entries, mark invalid as damaged and count gaps ("fragmentation"). If
	 * only gaps are found, the file is considered as "virgin" (no entries).
	 *
	 * @return	void
	 */
	public function analyzeFile () {
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] CALLED!', __METHOD__, __LINE__));

		// Make sure the file is initialized
		assert($this->isFileInitialized());

		// Init counters and gaps array
		$this->initCountersGapsArray();

		// Output message (as this may take some time)
		self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] Analyzing file structure ... (this may take some time)', __METHOD__, __LINE__));

		// First rewind to the begining
		$this->rewind();

		// Then try to load all entries
		while ($this->valid()) {
			// Go to next entry
			$this->next();

			// Get current entry
			$current = $this->getCurrentBlock();

			// Simply output it
			self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] current=%s', __METHOD__, __LINE__, print_r($current, TRUE)));
		} // END - while

		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput(sprintf('[%s:%d:] EXIT!', __METHOD__, __LINE__));
	}

	/**
	 * Advances to next "block" of bytes
	 *
	 * @return	void
	 * @todo	This method will load large but empty files in a whole
	 */
	public function next () {
		// Is there nothing to read?
		if (!$this->valid()) {
			// Nothing to read
			return;
		} // END - if

		// Make sure the block instance is set
		assert($this->getBlockInstance() instanceof CalculatableBlock);

		// First calculate minimum block length
		$length = $this->getBlockInstance()->caluclateMinimumBlockLength();

		// Short be more than zero!
		assert($length > 0);

		// Wait until a entry/block separator has been found
		$data = $this->getBackBuffer();
		while ((!$this->isEndOfFileReached()) && (!$this->getBlockInstance()->isBlockSeparatorFound($data))) {
			// Then read the block
			$data .= $this->read($length);
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('data()=' . strlen($data));
		} // END - if

		// EOF reached?
		if ($this->isEndOfFileReached()) {
			// Set whole data as current block
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('Calling setCurrentBlock(' . strlen($data) . ') ...');
			$this->setCurrentBlock($data);

			// Then abort here silently
			//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('EOF reached.');
			return;
		} // END - if

		/*
		 * Init back-buffer which is the data that has been found beyond the
		 * separator.
		 */
		$this->initBackBuffer();

		// Separate data
		$dataArray = explode(self::getBlockSeparator(), $data);

		// This array must contain two elements
		//* NOISY-DEBUG: */ self::createDebugInstance(__CLASS__)->debugOutput('dataArray=' . print_r($dataArray, TRUE));
		assert(count($dataArray) == 2);

		// Left part is the actual block, right one the back-buffer data
		$this->setCurrentBlock($dataArray[0]);
		$this->setBackBuffer($dataArray[1]);
	}

	/**
	 * Checks wether the current entry is valid (not at the end of the file).
	 * This method will return TRUE if an emptied (nulled) entry has been found.
	 *
	 * @return	$isValid	Whether the next entry is valid
	 */
	public function valid () {
		// Make sure the block instance is set
		assert($this->getBlockInstance() instanceof CalculatableBlock);

		// First calculate minimum block length
		$length = $this->getBlockInstance()->caluclateMinimumBlockLength();

		// Short be more than zero!
		assert($length > 0);

		// Get current seek position
		$seekPosition = $this->key();

		// Then try to read it
		$data = $this->read($length);

		// If some bytes could be read, all is fine
		$isValid = ((is_string($data)) && (strlen($data) > 0));

		// Get header size
		$headerSize = $this->getBlockInstance()->getHeaderSize();

		// Is the seek position at or beyond the header?
		if ($seekPosition >= $headerSize) {
			// Seek back to old position
			$this->seek($seekPosition);
		} else {
			// Seek directly behind the header
			$this->seek($headerSize);
		}

		// Return result
		return $isValid;
	}

	/**
	 * Gets current seek position ("key").
	 *
	 * @return	$key	Current key in iteration
	 */
	public function key () {
		// Call pointer instance
		return $this->getPointerInstance()->determineSeekPosition();
	}
}

// [EOF]
?>
